"""Wrapper around a subset of the multiprocessing built-in library
to make running multiple processes easy. Load balancing is left to
the caller."""

import multiprocessing

CPUS = multiprocessing.cpu_count()


def parallelize(inputs, cpus, func):
    """Parallelize some task on an input list across the specified
    number of CPU cores.

    :param inputs: arguments to call `func` with
    :type inputs: list
    :param cpus: number of CPU cores to use
    :type cpus: int
    :param func: function to call in parallel
    :type func: function
    :return: results
    """
    # Return early if called with no inputs
    if len(inputs) == 0:
        return list()

    # User requested some number of cores - make sure it's sane
    if cpus < 1 or cpus > CPUS:
        cpus = min([CPUS, len(inputs)])

    tasks = []
    for item in inputs:
        if not isinstance(item, tuple):
            item = (item,)
        tasks.append((func, item))

    return start_processes(tasks, cpus)


def worker(input_queue, output_queue):
    """Worker function to run jobs from the input queue until a stop
    signal is reached.

    :param input_queue: thread-safe queue to pull jobs from
    :type input_queue: multiprocessing.Queue
    :param output_queue: thread-safe queue to add results to
    :type output_queue: multiprocessing.Queue
    """
    for func, args in iter(input_queue.get, "STOP"):
        result = func(*args)
        if result:
            output_queue.put(result)


def start_processes(inputs, cpus):
    """Spool up processes and use them to run the jobs.

    :param inputs:
    :type inputs:
    :param cpus:
    :type cpus:
    :return: results
    """
    job_q = multiprocessing.Queue()
    done_q = multiprocessing.Queue()

    [job_q.put(job) for job in inputs]

    worker_pool = list()
    for i in range(cpus):
        worker_pool.append(
            multiprocessing.Process(target=worker, args=(job_q, done_q)))
        job_q.put("STOP")

    [w.start() for w in worker_pool]

    # Grab results
    results = []
    for _ in range(len(inputs)):
        results.append(done_q.get())

    [w.join() for w in worker_pool]

    return results



"""Functions for calculating various similarity metrics between pairs
of genomes."""

from phamclust.Genome import Genome
from phamclust.statistics import average


def percentage_of_conserved_proteins(source, target):
    """Calculate the percentage of conserved proteins between the
    given source and target genomes.

    Qin et al. A Proposed Genus Boundary for the Prokaryotes Based on
    Genomic Insights.
    Journal of Bacteriology (2014) Volume 196, Issue 12.
    DOI: https://doi.org/10.1128/JB.01688-14.

    :param source: the source genome to be compared
    :type source: Genome
    :param target: the target genome to be compared
    :type target: Genome
    :return: similarity
    """
    shared_phams = source.intersection(target)

    s_conserved = sum([source[x] for x in shared_phams])
    t_conserved = sum([target[x] for x in shared_phams])
    s_total = len(source)
    t_total = len(target)

    similarity = float(s_conserved + t_conserved) / (s_total + t_total)

    return round(similarity * 100, 2)


def gene_content_similarity(source, target):
    """Calculate the gene content similarity between the given source
    and target genomes.

    Mavrich and Hatfull. Bacteriophage Evolution Differs by Host,
    Lifestyle, and Genome.
    Nature Microbiology (2017) Volume 2.
    DOI: https://doi.org/10.1038/nmicrobiol.2017.112.

    :param source: the source genome to be compared
    :type source: Genome
    :param target: the target genome to be compared
    :type target: Genome
    :return: similarity
    """
    shared_phams = source.intersection(target)

    source_gcs = float(len(shared_phams)) / len(source)
    target_gcs = float(len(shared_phams)) / len(target)

    similarity = average([source_gcs, target_gcs])

    return round(similarity * 100, 2)


def jaccard_similarity_coefficient(source, target):
    """Calculate the jaccard similarity coefficient between the given
    source and target genomes.

    Jaccard. The Distribution of the Flora in the Alpine Zone.
    New Phytologist (1912) Volume 11, Issue 2.
    DOI: https://doi.org/10.1111/j.1469-8137.1912.tb05611.x.

    :param source: the source genome to be compared
    :type source: Genome
    :param target: the target genome to be compared
    :type target: Genome
    :return: similarity
    """
    shared_phams = source.intersection(target)
    all_phams = source.union(target)

    similarity = float(len(shared_phams)) / len(all_phams)

    return round(similarity * 100, 2)



from phamclust.multiprocess import parallelize
from phamclust.statistics import average, skewness, standard_deviation


class SymMatrix:
    """A SymMatrix is an undirected graph for storing pairwise edge
    data.

    This class should NOT be used for directed graphs - implementation
    does not preserve directionality of relationships.
    """
    def __init__(self, names):
        """Constructor for SymMatrix instances.

        :param names: node names for this matrix
        :type names: list[str]
        """
        self._node_names = sorted(names)
        self._matrix = dict()

        # Populate source names for O(1) membership testing
        for name in self._node_names:
            self._matrix[name] = dict()

        self.fill_diagonal(100.0)

    @property
    def node_names(self):
        """Return the node names."""
        # return self._matrix.keys()
        return self._node_names[:]

    @property
    def medoid(self):
        """Return the name of the most central node. The medoid node
        has the highest average similarity to all other nodes."""
        medoid, medoid_sum = None, 0
        for source in self._node_names:
            source_sum = sum(self[source].values())
            if source_sum > medoid_sum:
                medoid = source
                medoid_sum = source_sum

        return medoid

    @property
    def epsilon(self):
        """Return the matrix's skew-adjusted mean edge value."""
        if len(self) == 1:
            return 100.0

        edges = list()
        for _, __, weight in self:
            edges.append(weight)

        mean = average(edges)
        std_dev = standard_deviation(edges, mean)
        skew = skewness(edges, mean)

        return mean + (skew * std_dev)

    def reorder(self, names):
        """Use the incoming names as the row order for this matrix.

        :param names:
        :type names: list[str]
        """
        if len(names) != len(self):
            raise ValueError(f"expected len(names) == {len(self)}")

        for name in names:
            if name not in self:
                raise KeyError(f"node ({name}) not in matrix")

        self._node_names = names

    def iterrows(self):
        """Iterate over the matrix rows."""
        pass

    def extract_submatrix(self, names):
        """Return the sub-graph indicated by `names`.

        :param names: names of the nodes to build a submatrix from
        :type names: list[str]
        :return: submatrix
        """
        submatrix = SymMatrix(sorted(names))
        node_names = submatrix.node_names
        for i, source in enumerate(node_names):
            for target in node_names[i+1:]:
                weight = self.get_cell(source, target)
                submatrix.fill_cell(source, target, weight)

        return submatrix

    def nearest_neighbors(self, source, threshold):
        """Return the source node's nearest neighbors above threshold.

        :param source: name of the source node to search
        :type source: str
        :param threshold: similarity threshold to be neighbors
        :type threshold: float
        :return:
        """
        if source not in self:
            raise KeyError(f"node ({source}) not in matrix")

        neighbors = list()
        for target, weight in self[source].items():
            if weight >= threshold and source != target:
                neighbors.append(target)

        return neighbors

    def fill_diagonal(self, value):
        """Fill the matrix diagonal."""
        for source in self._node_names:
            self.fill_cell(source, source, value)

    def fill_cell(self, source, target, value):
        """Fill a specific cell of the matrix."""
        if source not in self:
            raise KeyError(f"source node ({source}) not in matrix")
        if target not in self:
            raise KeyError(f"target node ({target}) not in matrix")

        if not 0 <= value <= 100:
            raise ValueError(f"value ({value}) not in range(0, 100)")

        self._matrix[source][target] = value
        if source != target:
            self._matrix[target][source] = value

    def get_cell(self, source, target):
        """Return a specific cell of the matrix"""
        if source not in self:
            raise KeyError(f"source node ({source}) not in matrix")
        if target not in self:
            raise KeyError(f"target node ({target}) not in matrix")

        return self[source][target]

    def write_clustalo(self, filepath):
        """Write matrix to a file, as a Clustal Omega identity matrix.

        :param filepath: path to the file to write matrix to
        :type filepath: pathlib.Path
        """
        with open(filepath, "w") as matrix_writer:
            matrix_writer.write(str(self))

    def write_adjacency(self, filepath):
        """Write matrix to a file, as an adjacency matrix.

        :param filepath: path to the file to write matrix to
        :type filepath: pathlib.Path
        """
        with open(filepath, "w") as matrix_writer:
            for source, target, weight in self:
                matrix_writer.write(f"{source}\t{target}\t{weight}\n")

    def __str__(self):
        s = f"{len(self)}\n"
        for source in self._node_names:
            s += f"{source:<24}"
            for target in self._node_names:
                s += " " + str(self.get_cell(source, target))
            s += "\n"

        return s

    def __lt__(self, other):
        if not isinstance(other, SymMatrix):
            raise TypeError(f"other expected 'SymMatrix', not '{type(other)}'")

        return len(self) < len(other)

    def __getitem__(self, item):
        if not isinstance(item, str):
            raise TypeError(f"item expected 'str', not '{type(item)}'")

        return self._matrix[item]

    def __contains__(self, item):
        if not isinstance(item, str):
            raise TypeError(f"item expected 'str', not '{type(item)}'")

        return item in self._matrix

    def __iter__(self):
        """Iterate matrix as an adjacency matrix."""
        for i, source in enumerate(self._node_names):
            for target in self._node_names[i:]:
                yield source, target, self._matrix[source][target]

    def __len__(self):
        return len(self._matrix)


def matrix_from_adjacency(genomes, filepath):
    """Create a symmetric matrix by reading in an adjacency matrix
    file.

    :param genomes: the genomes to build a matrix from
    :type genomes: dict[str, PhamClust.Genome.Genome]
    :param filepath: path to an adjacency matrix file
    :type filepath: pathlib.Path
    :return: matrix
    """
    matrix = SymMatrix(list(genomes.keys()))

    with open(filepath, "r") as adjacency_reader:
        for row in adjacency_reader:
            source, target, value = row.rstrip().split()
            matrix.fill_cell(source, target, float(value))

    return matrix


# Helper function for matrix_de_novo() --> facilitates parallelization
def _calculate_adjacency(genomes, sources, targets, func):
    """Calculate the similarity between source and target genomes using
    `func`.

    :param genomes: all genomes in the dataset
    :type genomes: dict[str, PhamClust.Genome.Genome]
    :param sources: source genome list
    :type sources: list[str]
    :param targets: target genome lists
    :type targets: list[list[str]]
    :param func: similarity function to use for genome comparisons
    :type func: function
    """
    if len(sources) != len(targets):
        raise ValueError(f"expected len(sources) == len(targets)")

    adjacency = list()
    for source_name, target_list in zip(sources, targets):
        source_genome = genomes[source_name]
        for target_name in target_list:
            target_genome = genomes[target_name]
            weight = func(source_genome, target_genome)
            adjacency.append((source_name, target_name, weight))

    return adjacency


def matrix_de_novo(genomes, func, cpus, batch_size=20):
    """Create a symmetric matrix from scratch using `func` to measure
    the similarity between genomes.

    :param genomes: the genomes to build a matrix from
    :type genomes: dict[str, PhamClust.Genome.Genome]
    :param func: similarity function to use for genome comparisons
    :type func: function
    :param cpus: number of CPU cores to use for matrix-building
    :type cpus: int
    :param batch_size: approximate number of genomes per
    :type batch_size: int
    :return: matrix
    """
    matrix = SymMatrix(list(genomes.keys()))
    node_names = matrix.node_names

    num_batches = len(genomes) // batch_size
    while len(genomes) / num_batches > batch_size or num_batches % cpus != 0:
        num_batches += 1

    jobs = list()
    for batch_index in range(num_batches):
        batch_indices = range(len(node_names))[batch_index::num_batches]
        batch_sources = [node_names[x] for x in batch_indices]
        batch_targets = [node_names[x+1:] for x in batch_indices]
        jobs.append((genomes, batch_sources, batch_targets, func))

    for batch_result in parallelize(jobs, cpus, _calculate_adjacency):
        for source, target, value in batch_result:
            matrix.fill_cell(source, target, value)

    return matrix



"""Functions for clustering the genomes in a symmetric matrix."""

import queue

from phamclust.SymMatrix import SymMatrix


def single_linkage(matrix, eps=None):
    """Cluster matrix with single-linkage to the indicated threshold.

    Implementation uses dbscan with minpts=1.

    :param matrix: the matrix whose nodes need clustering
    :type matrix: SymMatrix
    :param eps: similarity threshold for clustering
    :type eps: float
    :return: clusters
    """
    return dbscan(matrix, eps, minpts=1)


def hierarchical(matrix, eps=None):
    """Cluster matrix hierarchically to the indicated threshold.

    Implementation uses dbscan with minpts=2.

    :param matrix: the matrix whose nodes need clustering
    :type matrix: SymMatrix
    :param eps: similarity threshold for clustering
    :type eps: float
    :return: clusters
    """
    return dbscan(matrix, eps, minpts=2)


def dbscan(matrix, eps=None, minpts=3):
    """Cluster matrix with dbscan to the indicated threshold.

    :param matrix: the matrix whose nodes need clustering
    :type matrix: SymMatrix
    :param eps: similarity threshold for clustering
    :type eps: float
    :param minpts: minimum number of neighbors to build a cluster
    :type minpts: int
    :return: clusters
    """
    if not eps:
        eps = matrix.epsilon

    counter = 0
    cluster_lookup = dict()
    for source_node in matrix.node_names:
        # Skip nodes we clustered already
        if cluster_lookup.get(source_node, -1) != -1:
            continue

        neighbors = matrix.nearest_neighbors(source_node, eps)
        if len(neighbors) < minpts:
            cluster_lookup[source_node] = None
            continue

        counter += 1
        cluster_lookup[source_node] = counter

        seeds = queue.Queue()
        for neighbor in neighbors:
            seeds.put(neighbor)

        while not seeds.empty():
            target = seeds.get()

            target_cluster = cluster_lookup.get(target, -1)
            if not target_cluster:
                cluster_lookup[target] = counter
            if target_cluster != -1:
                continue
            cluster_lookup[target] = counter
            neighbors = matrix.nearest_neighbors(target, eps)

            if len(neighbors) >= minpts:
                for neighbor in neighbors:
                    seeds.put(neighbor)

    clusters = dict()
    for node_name, cluster in cluster_lookup.items():
        if not cluster:
            counter += 1
            clusters[counter] = [node_name]
        else:
            cluster_nodes = clusters.get(cluster, list())
            cluster_nodes.append(node_name)
            clusters[cluster] = cluster_nodes
    for cluster, cluster_nodes in clusters.items():
        cluster_matrix = matrix.extract_submatrix(cluster_nodes)
        edges = cluster_matrix[cluster_matrix.medoid]
        edges = sorted(edges, key=lambda x: edges[x], reverse=True)
        cluster_matrix.reorder(edges)
        clusters[cluster] = cluster_matrix

    return sorted([c for c in clusters.values()], reverse=True)


def lloyds(matrix, medoids, eps=None, max_iterations=100):
    """Cluster the matrix using Lloyd's algorithm to the indicated
    threshold.

    :param matrix:
    :type matrix: SymMatrix
    :param medoids:
    :type medoids:
    :param eps:
    :type eps: float
    :param max_iterations:
    :type max_iterations: int
    :return: clusters
    """
    pass



"""A program for assorting genomes into clusters and sub-clusters
based on shared gene content. Proteins evolve more slowly than the DNA
that encodes them, making this approach much more robust for comparing
the relatedness of bacteria between genera, or phages at almost any
genetic distance."""

import sys
import argparse
from pathlib import Path

from phamclust import clustering, similarity, multiprocess as mp
from phamclust.Genome import Genome
from phamclust.SymMatrix import matrix_de_novo

metrics = {"pocp": similarity.percentage_of_conserved_proteins,
           "gcs": similarity.gene_content_similarity,
           "jaccard": similarity.jaccard_similarity_coefficient}

CWD = Path().cwd()
METRIC = "pocp"
THRESH = 35


def parse_args():
    """Parse commandline arguments."""
    p = argparse.ArgumentParser(description=__doc__, prog="phamclust")

    p.add_argument("infile", type=Path,
                   help="path to tab-delimited file containing genome-pham "
                        "mappings")
    p.add_argument("outdir", type=Path, default=CWD,
                   help=f"directory where output files should be written "
                        f"[default: {CWD}]")

    # Clustering parameters
    c = p.add_argument_group("clustering parameters (optional)")
    c.add_argument("-m", "--metric", default=METRIC, type=str,
                   choices=metrics.keys(), metavar="",
                   help=f"metric to use for calculating pairwise similarity "
                        f"[default: {METRIC}]")
    c.add_argument("-t", "--threshold", default=THRESH, type=float, metavar="",
                   help=f"similarity threshold for a genome to join a cluster "
                        f"[default: {THRESH}%%]")
    c.add_argument("-s", "--subcluster", action="store_true",
                   help="break clusters into subclusters")

    p.add_argument("-c", "--cpus", default=mp.CPUS, type=int, metavar="",
                   help=f"number of CPU cores to use [default: {mp.CPUS}]")

    return p.parse_args()


def build_genomes(infile):
    """Parse genome-pham edges from the input file and populate a
    symmetric matrix with pairwise genome similarities.

    :param infile: tab-delimited file containing genome-pham edges
    :type infile: Path
    :return: matrix
    """
    genomes = dict()
    with open(infile, "r") as edge_reader:
        for row in edge_reader:
            name, pham = row.rstrip().split()
            if name not in genomes:
                genomes[name] = Genome(name)

            genomes[name].add(pham)

    return genomes


def get_subclusters(clusters):
    subclusters = [[]] * len(clusters)

    for i, cluster in enumerate(clusters):
        # If cluster too small to robustly determine eps, continue
        if len(cluster) < 5:
            continue

        # Otherwise - check eps to decide whether to subdivide
        eps = max([50, round(cluster.epsilon)])
        if eps >= 90:
            continue

        subclusters[i].extend(clustering.dbscan(cluster, eps))
        cluster_order = list()
        for j, subcluster in enumerate(subclusters[i]):
            cluster_order.extend(subcluster.node_names)
        cluster.reorder(cluster_order)

    return subclusters


def get_clusters(infile, threshold, func=metrics[METRIC], cpus=1):
    """Assemble clusters and subclusters.

    :param infile: tab-delimited file containing genome-pham edges
    :type infile: Path
    :param threshold: threshold to use for clustering
    :type threshold: int
    :param func: function to use for calculating similarity
    :type func: function
    :param cpus: number of CPU cores to use
    :type cpus: int
    :return: clusters
    """
    genomes = build_genomes(infile)

    matrix = matrix_de_novo(genomes, func, cpus)

    return clustering.dbscan(matrix, threshold)


def main():
    """Commandline interface for this module."""
    if len(sys.argv) == 1:
        sys.argv.append("-h")

    namespace = parse_args()

    infile = namespace.infile
    if not infile.is_file():
        print(f"input file ({infile}) is not a valid file - exiting...")
        sys.exit(1)

    outdir = namespace.outdir
    if not outdir.is_dir():
        print(f"output directory ({outdir}) does not exist - making it...")
        outdir.mkdir(parents=True)
    else:
        print(f"output directory ({outdir}) exists - overwriting contents...")
        response = input(f"press 'Enter' to continue, any other key to abort")
        if response:
            print(f"received abort signal - exiting...")
            sys.exit(1)

    metric = metrics[namespace.metric]
    threshold = namespace.threshold
    cpus = namespace.cpus
    do_sub = namespace.subcluster

    clusters = get_clusters(infile, threshold, metric, cpus)
    if do_sub:
        subclusters = get_subclusters(clusters)
    else:
        subclusters = [[]] * len(clusters)

    for i, group in enumerate(zip(clusters, subclusters)):
        _cluster, _subclusters = group

        _cluster.write_clustalo(outdir.joinpath(f"cluster_{i+1}.txt"))
        if not _subclusters or len(_subclusters) == 1:
            continue

        for j, _subcluster in enumerate(_subclusters):
            outfile = outdir.joinpath(f"cluster_{i+1}.{j+1}.txt")
            _subcluster.write_clustalo(outfile)


if __name__ == "__main__":
    main()



"""A dictionary-based class for efficiently storing the gene content
encoded by genomes, and intuitively comparing genomes to each other."""


class Genome:
    """A Genome is a collection of gene phamilies that may or may not
    be shared by other Genomes. Paralogs (multiple copies of the same
    phamily) are tolerated.

    Class methods allow this object to behave like something between a
    set and a dictionary."""
    def __init__(self, name, phams=None):
        """Constructor for Genome instances.

        :param name: name for this genome
        :type name: str
        :param phams: phamilies found in this genome
        :type phams: list[str]
        """
        self.name = name
        self.phams = dict()

        if phams:
            if not isinstance(phams, list):
                raise TypeError(f"phams expected 'list', not '{type(phams)}'")

            for pham in phams:
                self.add(pham)

    def add(self, pham):
        """Add a pham to this genome.

        :param pham: name of the pham to add to this genome
        :type pham: str
        """
        if not isinstance(pham, str):
            raise TypeError(f"pham expected 'str', not '{type(pham)}'")

        if pham not in self:
            self.phams[pham] = 1
        else:
            self.phams[pham] += 1

    def pop(self, pham):
        """Pop a pham from this genome.

        :param pham: name of the pham to pop from this genome
        :type pham: str
        """
        if not isinstance(pham, str):
            raise TypeError(f"pham expected 'str', not '{type(pham)}'")

        self.phams.pop(pham)

    def intersection(self, other):
        """Get the set of phams present both in this Genome and
        another Genome.

        :param other: another genome to be checked for shared phams
        :type other: Genome
        """
        if not isinstance(other, Genome):
            raise TypeError(f"other expected 'Genome', not '{type(other)}'")

        return set(self.phams.keys()).intersection(other.phams.keys())

    def union(self, other):
        """Get the set of phams present in either this Genome or
        another Genome.

        :param other: another genome to be checked for shared phams
        :type other: Genome
        """
        if not isinstance(other, Genome):
            raise TypeError(f"other expected 'Genome', not '{type(other)}'")

        return set(self.phams.keys()).union(other.phams.keys())

    def difference(self, other):
        """Get the set of phams present in this Genome but not another
        Genome.

        :param other: another genome to be checked for shared phams
        :type other: Genome
        """
        if not isinstance(other, Genome):
            raise TypeError(f"other expected 'Genome', not '{type(other)}'")

        return set(self.phams.keys()).difference(other.phams.keys())

    def symmetric_difference(self, other):
        """Get the set of phams present in either this Genome or
        another, but not both.

        :param other: another genome to be checked for shared phams
        :type other: Genome
        """
        if not isinstance(other, Genome):
            raise TypeError(f"other expected 'Genome', not '{type(other)}'")

        return set(self.phams.keys()).symmetric_difference(other.phams.keys())

    def __lt__(self, other):
        if not isinstance(other, Genome):
            raise TypeError(f"other expected 'Genome', not '{type(other)}'")

        return len(self) < len(other)

    def __getitem__(self, item):
        if not isinstance(item, str):
            raise TypeError(f"item expected 'str', not '{type(item)}'")

        return self.phams[item]

    def __contains__(self, item):
        if not isinstance(item, str):
            raise TypeError(f"item expected 'str', not '{type(item)}'")

        return item in self.phams

    def __iter__(self):
        for key, value in self.phams.items():
            yield key, value

    def __len__(self):
        return sum([value for key, value in self])



"""Functions to calculate summary statistics."""


def average(values):
    """Calculate the arithmetic mean of the given values.

    :param values: values to average
    :type values: list[float or int]
    :return: average
    """
    return float(sum(values)) / len(values)


def variance(values, mean=None, sample=True):
    """Calculate the variance of the given values.

    :param values: values to get variance of
    :type values: list[float or int]
    :param mean: the mean of the given values
    :type mean: float
    :param sample: sample or whole population?
    :type sample: bool
    :return: variance
    """
    if len(values) == 1:
        return 0.0

    if not mean:
        mean = average(values)

    numerator = sum([(x - mean) ** 2 for x in values])

    if sample:
        denominator = len(values) - 1
    else:
        denominator = len(values)

    return numerator / denominator


def standard_deviation(values, mean=None, sample=True):
    """Calculate the standard deviation of the given values.

    :param values: values to get variance of
    :type values: list[float or int]
    :param mean: the mean of the given values
    :type mean: float
    :param sample: sample or whole population?
    :type sample: bool
    :return: std_dev
    """
    return variance(values, mean, sample) ** 0.5


def skewness(values, mean=None, sample=True):
    """Calculate the skewness of the given values.

    :param values: values to get variance of
    :type values: list[float or int]
    :param mean: the mean of the given values
    :type mean: float
    :param sample: sample or whole population?
    :type sample: bool
    :return: skewness
    """
    if len(values) == 1:
        return 0.0

    if not mean:
        mean = average(values)

    numerator = sum([(x - mean) ** 3 for x in values])
    denominator = standard_deviation(values, mean, sample) ** 3

    if sample:
        denominator *= (len(values) - 1)
    else:
        denominator *= len(values)

    return numerator / denominator
